@@ -8,7 +8,6 @@
 
 from __future__ import absolute_import
 
-import json
 import logging
 import logging.config
 import os
@@ -25,6 +24,7 @@ import webob
 
 from webob.exc import (
     HTTPBadRequest,
+    HTTPForbidden,
     HTTPNotFound,
 )
 
@@ -34,6 +34,7 @@ from ovirt_imageio_common import errors
 from ovirt_imageio_common import ssl
 from ovirt_imageio_common import util
 from ovirt_imageio_common import version
+from ovirt_imageio_common import validate
 from ovirt_imageio_common import web
 
 from . import config
@@ -127,15 +128,6 @@ def start_server(config, server, name):
                       name=name)
 
 
-def response(status=200, payload=None):
-    """
-    Return WSGI application for sending response in JSON format.
-    """
-    body = json.dumps(payload) if payload else ""
-    return webob.Response(status=status, body=body,
-                          content_type="application/json")
-
-
 class Images(object):
     """
     Request handler for the /images/ resource.
@@ -156,21 +148,29 @@ class Images(object):
             raise HTTPBadRequest("Invalid Content-Length header: %r" % size)
         content_range = web.content_range(self.request)
         offset = content_range.start or 0
+
+        # For backward compatibility, we flush by default.
+        flush = validate.enum(self.request.params, "flush", ("y", "n"),
+                              default="y")
+        flush = (flush == "y")
+
         ticket = tickets.authorize(ticket_id, "write", offset + size)
         # TODO: cancel copy if ticket expired or revoked
-        self.log.info("Writing %d bytes at offset %d to %s for ticket %s",
-                      size, offset, ticket.url.path, ticket_id)
+        self.log.info(
+            "Writing %d bytes at offset %d flush %s to %s for ticket %s",
+            size, offset, flush, ticket.url.path, ticket_id)
         op = directio.Receive(ticket.url.path,
                               self.request.body_file_raw,
                               size,
                               offset=offset,
+                              flush=flush,
                               buffersize=self.config.daemon.buffer_size)
         ticket.add_operation(op)
         try:
             op.run()
         except errors.PartialContent as e:
             raise HTTPBadRequest(str(e))
-        return response()
+        return web.response()
 
     def get(self, ticket_id):
         # TODO: cancel copy if ticket expired or revoked
@@ -216,6 +216,76 @@ class Images(object):
 
         return resp
 
+    def patch(self, ticket_id):
+        if not ticket_id:
+            raise HTTPBadRequest("Ticket id is required")
+        try:
+            msg = self.request.json
+        except ValueError as e:
+            raise HTTPBadRequest("Invalid JSON message: %s" % e)
+
+        op = validate.enum(msg, "op", ("zero", "flush"))
+        if op == "zero":
+            return self._zero(ticket_id, msg)
+        elif op == "flush":
+            return self._flush(ticket_id, msg)
+        else:
+            raise RuntimeError("Unreachable")
+
+    def _zero(self, ticket_id, msg):
+        size = validate.integer(msg, "size", minval=0)
+        offset = validate.integer(msg, "offset", minval=0, default=0)
+        flush = validate.boolean(msg, "flush", default=False)
+
+        ticket = tickets.authorize(ticket_id, "write", offset + size)
+
+        self.log.info(
+            "Zeroing %d bytes at offset %d flush %s to %s for ticket %s",
+            size, offset, flush, ticket.url.path, ticket_id)
+        op = directio.Zero(ticket.url.path, size, offset=offset, flush=flush,
+                           buffersize=self.config.daemon.buffer_size)
+        ticket.add_operation(op)
+        try:
+            op.run()
+        except errors.PartialContent as e:
+            raise HTTPBadRequest(str(e))
+        return web.response()
+
+    def _flush(self, ticket_id, msg):
+        ticket = tickets.authorize(ticket_id, "write", 0)
+        self.log.info("Flushing %s for ticket %s", ticket.url.path, ticket_id)
+        op = directio.Flush(ticket.url.path)
+        ticket.add_operation(op)
+        op.run()
+        return web.response()
+
+    def options(self, ticket_id):
+        if not ticket_id:
+            raise HTTPBadRequest("Ticket id is required")
+
+        if ticket_id == "*":
+            # Reporting the meta-capabilities for all images.
+            allow = ["OPTIONS", "GET", "PUT", "PATCH"]
+            features = ["zero", "flush"]
+        else:
+            # Reporting real image capabilities per ticket.
+            try:
+                ticket = tickets.get(ticket_id)
+            except KeyError:
+                raise HTTPForbidden("No such ticket %r" % ticket_id)
+
+            allow = ["OPTIONS"]
+            features = []
+            if ticket.may("read"):
+                allow.append("GET")
+            if ticket.may("write"):
+                allow.extend(("PUT", "PATCH"))
+                features = ["zero", "flush"]
+
+        return web.response(
+            payload={"features": features},
+            allow=",".join(allow))
+
 
 class Tickets(object):
     """
@@ -235,7 +305,7 @@ class Tickets(object):
         except KeyError:
             raise HTTPNotFound("No such ticket %r" % ticket_id)
         self.log.info("Retrieving ticket %s", ticket_id)
-        return response(payload=ticket.info())
+        return web.response(payload=ticket.info())
 
     def put(self, ticket_id):
         # TODO
@@ -255,7 +325,7 @@ class Tickets(object):
             raise HTTPBadRequest("Invalid ticket: %s" % e)
 
         tickets.add(ticket_id, ticket)
-        return response()
+        return web.response()
 
     def patch(self, ticket_id):
         # TODO: restart expire timer
@@ -278,7 +348,7 @@ class Tickets(object):
         except KeyError:
             raise HTTPNotFound("No such ticket: %s" % ticket_id)
         ticket.extend(timeout)
-        return response()
+        return web.response()
 
     def delete(self, ticket_id):
         # TODO: cancel requests using deleted tickets
@@ -289,7 +359,7 @@ class Tickets(object):
                 raise HTTPNotFound("No such ticket %r" % ticket_id)
         else:
             tickets.clear()
-        return response(status=204)
+        return web.response(status=204)
 
 
 class ThreadedWSGIServer(socketserver.ThreadingMixIn,
